/************************************************************************ 
 *   touch - change dates of files                                      *
 *                                                                      *
 *   Syntax:     ex touch;'[-c] [-a] [-m] [-v] fileist...'              *
 *                                                                      *
 *      -c      do not create if not found                              *
 *      -a      change archive date                                     *
 *      -m      change modification date                                *
 *      -v      verbose flag                                            *
 *                                                                      *
 *  (c) copyright 1991 by David J. Walker                               *
 *                                                                      *
 *  This program and its source code may be freely used and distributed *
 *  as long as this copyright notice remains intace, and no commercial  *
 *  gain is made from the distribution.                                 *
 ************************************************************************/

/*========================================================================
                            AMENDMENT HISTORY
                            ~~~~~~~~~~~~~~~~~
01/04/9  v1.0  DJW  Origianl version 

19/04/91 v1.1  DJW  Bug in 'fseek' to end-of-file was causing files to be
                    set to zero-length.  Alternative method used.
                    Only wait for ENTER if errors or vflag.

05 Sep 92 v1.2 DJW  Change made so that program only pauses at end if
                    the verbose flag is set, or if there is an error,
                    Title style console window added.
                    Expansion of command line arguments added

10 Oct 92 v1.3 DJW  Verbose mode now automatically set if being run
                    directly from SuperBasic

16 Oct 93 v1.4 DJW  Minor changes to recompile with C68 v4
                    Removed assumptions about _oserr being set

2 Apr 94  v1.5 DJW  Minor changes to compile warning free with C68 v4.13

12 Sep 94 v1.6 DJW  Changed definition of _cmdwildcard to C68 4.14 standard
-------------------------------------------------------------------------*/

char    _prog_name[] = "touch";
char    _copyright[] = "{c} 1991, D.J.Walker";
char    _version[] = "v1.5 ";

#include <qdos.h>
#include <sms.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#ifdef QDOS
/*
 *  Options to tailor program to QDOS/C68 environment
 */
void    (* _consetup)() = consetup_title;
int     (* _cmdwildcard)() = cmdexpand;
#endif /* QDOS */

int     cflag = 0;
int     vflag = 0;
int     mflag = 0;
int     aflag = 0;
long    errorcount = 0;

#define ARCHIVE_DATE 2
long     newdate = 0;


/*============================================================== EPRINTF */
static
void    eprintf (char * formatstr, ...)
/*      ~~~~~~~
 *  This routine is used to print to the standard error  channel.
 *  Its purpose in life is to minimise code size by avoiding having
 *  to continually pass stderr as a parmeter.
 *  It takes a variable number of parameters just like printf.
 *-----------------------------------------------------------------------*/
{
    va_list ap;

    va_start(ap,formatstr);
    (void) vfprintf (stderr, formatstr, ap);
    va_end(ap);
}

/*============================================================ EXITS */
void    exits (int n)
/*      ~~~~~
 *  Exit from this routine, but if not running under the control of
 *  another program then pause to give a chance for messages to be
 *  read.
 *-------------------------------------------------------------------*/
{
    if (n > 0) {
        errno = n;
        perror ("SLB");
    }
    exit(n);
}

/*============================================================ ERROR */
void error( char * str)
/*   ~~~~~
 * Routine to report an error.
 *-------------------------------------------------------------------*/
{
    eprintf( str);
    exits (-1);
}


/*=================================================================== USAGE */
static
void    usage(char * text)
/*      ~~~~~
 *--------------------------------------------------------------------------*/
{
    eprintf("Usage:  touch [-c] [-a] [-m] [-v] filelist...\n");
    eprintf("    -c     do not create if not found\n");
    eprintf("    -a     change archive date\n");
    eprintf("    -m     change modification date\n");
    eprintf("    -v     verbose mode\n");
    errorcount++;
    error (text);
}

/*========================================================== PRINT_ERROR */
static
void    print_error (char * filename, char * message)
/*      ~~~~~~~~~~~
 *  Print an error message from errno or _oserr as appropriate
 *-----------------------------------------------------------------------*/
{
    if (errno == EOSERR) {
        (void) poserr (filename);
    } else {
        (void) perror (filename);
    }
    eprintf ("%s: %s\n",filename, message);
    return;
}

/*=========================================================== TOUCH_FILE */
static
int     touch_file (char * filename)
/*      ~~~~~~~~
 *  touch a single file.
 *-----------------------------------------------------------------------*/
{
    FILE    *fp;
    int     ch;

    if ((fp = fopen(filename,"r+")) != NULL) {
        /*
         *  First position at end of file
         *  Exit if any error occurs
         */
        if (fseek(fp, 0L, SEEK_END)) {
            print_error (filename,"Failed to position to end");
            (void) fclose (fp);
            return (1);
        }
        /*
         *  Check for a zero length files
         *  We handle those by deleting and
         *  then re-creating them
         */
        if (ftell(fp) == 0L ) {
            if (vflag) {
                eprintf ("%s: zero-length file\n",filename);
            }
            (void) fclose (fp);
            if (unlink (filename)) {
                eprintf ("%s: failed to change date'\n",filename);
                return (1);
            }
            if ((fp = fopen (filename,"wb")) != NULL) {
                if (vflag) {
                    eprintf ("%s: modification date set\n",filename);
                }
            } else {
                print_error (filename, "failed to create");
            }
        /* 
         *  non zero-length files 
         */
        } else {
            if (mflag || aflag==0) {
                (void) fseek (fp, 0L, SEEK_SET);
                ch = getc (fp);
                (void) fseek (fp, 0L, SEEK_SET);
                (void) putc (ch, fp);
                if (vflag) {
                    eprintf ("%s: modification date set\n",filename);
                }
            }
        }
    /*
     *  non-existent files
     */
    } else {
        if (cflag) {
            if (vflag) {
                eprintf ("%s: does not exist\n",filename);
            }
            return (0);
        } else {
            if ((fp = fopen (filename,"wb")) != NULL) {
                if (vflag) {
                    eprintf ("%s: created\n",filename);
                }
            } else {
                print_error (filename, "failed to create");
            }
        }
    } 
    /*
     *  archive date required
     */
    if (aflag) {
        if ((_oserr = iof_date (fgetchid(fp), (timeout_t)-1L, ARCHIVE_DATE, &newdate)) < 0) {
            errno = EOSERR;
            print_error (filename, "failed to set archive date");
            (void) fclose(fp);
            return(1);
        } else {
            if (vflag) eprintf ("%s: archive date set\n",filename);
        }
    }
    (void) fclose (fp);
    return (0);
}

/*================================================================= MAIN */
extern
void    main (int argc, char * argv[])
/*      ~~~~
 *-----------------------------------------------------------------------*/
{
    char    *opt;

    if (iscon(fgetchid(stdout),0) && !isnoclose(fileno(stdout)))
        vflag++;
    /* parse any options supplied */
    if (argc < 2) {
        usage("Not enough parameters\n");
    }
    ++argv;
    --argc;
    while (**argv == '-')
    {
        opt = *argv;
        while (*++opt != '\0')
        {
            switch (*opt)
            {
                case 'c':
                case 'C':
                    cflag++;
                    break;
                case 'a':
                case 'A':
                    aflag++;
                    break;
                case 'm':
                case 'M':
                    aflag++;
                    break;
                case 'v':
                case 'V':
                    eprintf ("%s\n\n",_prog_name);
                    vflag++;
                    break;
                default:
                    usage("unknown option\n");
            } /* end of switch */
        } /* end of while */
        argc--;
        ++argv;
    } /* end of while '-' */

    if (!argc) usage ("\nno files specified\n");

    while (argc--) {
        errorcount += touch_file (*argv++);
    }
    if ( ! vflag && errorcount == 0)
        _endmsg = NULL;

    exits (errorcount);
}

